from enum import IntEnum


# fmt: off
class Piki(IntEnum):
    NONE           = 0x00
    RED_LEFT       = 0x10
    RED_DOWN       = 0x11
    RED_RIGHT      = 0x12
    RED_UP         = 0x13
    BLUE_LEFT      = 0x20
    BLUE_DOWN      = 0x21
    BLUE_RIGHT     = 0x22
    BLUE_UP        = 0x23
    YELLOW_LEFT    = 0x30
    YELLOW_DOWN    = 0x31
    YELLOW_RIGHT   = 0x32
    YELLOW_UP      = 0x33
    BULBORB_LEFT   = 0x40
    BULBORB_DOWN   = 0x41
    BULBORB_RIGHT  = 0x42
    BULBORB_UP     = 0x43
    WOLLYHOP_LEFT  = 0x50
    WOLLYHOP_DOWN  = 0x51
    WOLLYHOP_RIGHT = 0x52
    WOLLYHOP_UP    = 0x53
# fmt: on



from .tile import Tile
from .piki import Piki
from .player import Player
from piket.constants import (
    PLUCKING_PIKMIN,
    PLUCKING_PIKMIN_HEADER_LENGTH as HEADER_LEN,
    PLUCKING_PIKMIN_LAYER_LENGTH as LAYER_LEN,
    PLUCKING_WIDTH as WIDTH,
    PLUCKING_HEIGHT as HEIGHT,
    PLUCKING_LAYERS as LAYERS,
)
from piket.base.level_base import LevelBase
from typing import Self


class Level(LevelBase):
    def __init__(
        self,
        index: int,
        tiles: bytearray = bytearray(WIDTH * HEIGHT),
        pikis: bytearray = bytearray(WIDTH * HEIGHT),
        raw: bytes | bytearray = bytearray(0),
        grid: tuple[int, int] = (0, 0),
        start: tuple[int, int] = (0, 0),
        player: Player = Player.OLIMAR,
    ):
        super().__init__(index, WIDTH, HEIGHT, LAYERS, [tiles, pikis], raw)
        self.grid = grid
        self.start = start
        self.player = player

    @classmethod
    def from_bytes(cls, level: bytes | bytearray) -> Self:
        level = bytearray(level)
        index = level[0]
        grid = level[1], level[2]
        start = level[3] % WIDTH, level[4] % HEIGHT
        player = Player.OLIMAR if level[3] < WIDTH else Player.LOUIE
        layers = level[HEADER_LEN:]
        tiles = layers[:LAYER_LEN]
        pikis = layers[LAYER_LEN : LAYER_LEN * 2]
        return cls(index, tiles, pikis, level, grid, start, player)

    def to_bytes(self) -> bytes:
        raw = bytearray()
        raw.extend(PLUCKING_PIKMIN)
        raw.append(self.index)
        raw.extend(self.grid)
        raw.append(self.start[0] + (self.width if self.player == Player.LOUIE else 0))
        raw.append(self.start[1] + (self.height if self.player == Player.LOUIE else 0))
        return super()._to_bytes(raw)

    def get_tile(self, x: int, y: int) -> Tile:
        """Gets the Tile at (x, y, layer 0)."""
        value = super()._get_tile(x, y, 0)
        if value not in Tile._value2member_map_:
            raise ValueError(f"Unknown Tile with value {value}")
        return Tile(value)

    def get_piki(self, x: int, y: int) -> Piki:
        """Gets the Piki at (x, y, layer 1)."""
        value = super()._get_tile(x, y, 2)
        if value not in Piki._value2member_map_:
            raise ValueError(f"Unknown Piki with value {value}")
        return Piki(value)

    def set_tile(self, x: int, y: int, tile: Tile | Piki):
        """Sets the (Tile | Piki) at (x, y) on the correct layer."""
        value = tile.value
        layer = self._enum_to_layer(tile)
        super()._set_tile(x, y, value, layer)

    def set_tiles(self, x: int, y: int, w: int, h: int, tile: Tile | Piki):
        """Sets the (Tiles | Pikis) from (x, y) to (w, h) on the correct layer."""
        value = tile.value
        layer = self._enum_to_layer(tile)
        super()._set_tiles(x, y, w, h, value, layer)

    def set_grid(self, tile: Tile | Piki):
        """Sets the (Tiles | Pikis) from (0, 0) to (grid_w, grid_h) on the correct layer."""
        value = tile.value
        layer = self._enum_to_layer(tile)
        super()._set_tiles(0, 0, self.grid[0], self.grid[1], value, layer)

    def clear_all(self):
        """Sets all Tiles and Pikis to value 0."""
        for i in range(self.layers):
            super()._set_tiles(0, 0, self.width, self.height, 0, i)

    def _enum_to_layer(self, tile: Tile | Piki) -> int:
        if isinstance(tile, Tile):
            return 0
        else:
            return 1



import logging
from pathlib import Path
from piket import NEDCENC, NEVPK
from piket.constants import SIZE_INFO, VPK_SIZE, VPK
from . import to_bytes, run_tool
from .headerfix import headerfix

logger = logging.getLogger(__file__)


def encode(
    data: bytes | bytearray | str | Path,
    original: bytes | bytearray | str | Path,
    partial_encode: bool = False,
) -> bytearray:
    PARENT = NEDCENC.parent.resolve()

    # handle all input types
    original = to_bytes(original)

    if len(original) > 0xA + 8 and original[0x1A : 0x1A + 8] == b"NINTENDO":
        decoded = bytearray(original)

    else:
        original_file = PARENT / "original.raw"
        logger.debug(f"Writing .raw original to '{original_file}'.")
        original_file.write_bytes(original)

        # decode original raw to use as template
        raw_decoded_path = PARENT / "raw_decoded.bin"
        logger.debug(f"Running nedcenc, output to '{raw_decoded_path}'.")
        run_tool(f'"{NEDCENC}" -i "{original_file}" -d -o "{raw_decoded_path}"')
        if not raw_decoded_path.exists():
            raise Exception("nedcenc did not output a file.")
        logger.debug(f"Removing '{original_file}'.")
        original_file.unlink()

        decoded = bytearray(raw_decoded_path.read_bytes())
        logger.debug(f"Removing '{raw_decoded_path}'.")
        raw_decoded_path.unlink()

    data = to_bytes(data)
    data_file = PARENT / "in.bin"
    logger.debug(f"Writing .bin data to '{data_file}'.")
    data_file.write_bytes(data)

    # compress new data with nevpk
    compressed_path = PARENT / "trimmed.vpk"
    logger.debug(f"Running nevpk, output to '{compressed_path}'.")
    run_tool(f'"{NEVPK}" -i "{data_file}" -c -o "{compressed_path}"')
    if not compressed_path.exists():
        raise Exception("nevpk did not output a file.")
    logger.debug(f"Removing '{data_file}'.")
    data_file.unlink()

    vpk = compressed_path.read_bytes()
    logger.debug(f"Removing '{compressed_path}'.")
    compressed_path.unlink()

    size = len(vpk)
    if size > 0xFFFF:
        raise ValueError(f"VPK size larger than maximum (expected <= 0xFFFF, got {size:#x}).")

    # re-write size info section
    logger.debug("Using template decoded data to build new deocded data.")
    size_info = int.from_bytes(decoded[SIZE_INFO : SIZE_INFO + 4], "little")
    size_info = (size_info & ~(0xFFFF << 9)) | ((size + 2) << 9)
    decoded[SIZE_INFO : SIZE_INFO + 4] = size_info.to_bytes(4, "little")
    # re-write vpk size
    decoded[VPK_SIZE : VPK_SIZE + 2] = size.to_bytes(2, "little")
    # re-write vpk
    for i in range(size):
        decoded[VPK + i] = vpk[i]
    # write incremenetal padding
    for i in range(VPK + size, 0x530):
        decoded[i] = (i - (VPK + size)) % 0x100

    decoded_path = PARENT / "decoded.bin"
    logger.debug(f"Writing rebuilt decoded data to '{decoded_path}'.")
    decoded_path.write_bytes(bytes(decoded))

    decoded = decoded_path.read_bytes()
    logger.debug(f"Removing '{decoded_path}'.")
    decoded_path.unlink()

    logger.debug("Running headerfix.")
    fixed = headerfix(decoded)

    if partial_encode:
        return bytearray(fixed)

    fixed_path = PARENT / "fixed.bin"
    fixed_path.write_bytes(fixed)

    raw_path = PARENT / "card.raw"
    logger.debug(f"Running nedcenc, output to '{raw_path}'.")
    run_tool(f'"{NEDCENC}" -i "{fixed_path}" -e -o "{raw_path}"')
    logger.debug(f"Removing '{fixed_path}'.")
    fixed_path.unlink()

    raw = raw_path.read_bytes()
    logger.debug(f"Removing '{raw_path}'.")
    raw_path.unlink()

    logger.info("Conversion from (.bin + .raw) to .raw complete.")
    return bytearray(raw)



from enum import IntEnum


# fmt: off
class Tile(IntEnum):
    GROUND          = 0x00
    FIRE            = 0x10
    WATER           = 0x11
    ELECTRICITY     = 0x12
    BULBORB         = 0x13
    POISON          = 0x14
    ROCK            = 0x20
    SOFT_ROCK_0     = 0x30 # never breaks
    SOFT_ROCK_1     = 0x31
    SOFT_ROCK_2     = 0x32
    SOFT_ROCK_3     = 0x33
    SOFT_ROCK_4     = 0x34
    SOFT_ROCK_5     = 0x35
    SOFT_ROCK_6     = 0x36
    SOFT_ROCK_7     = 0x37
    SOFT_ROCK_8     = 0x38
    SOFT_ROCK_9     = 0x39
    SOFT_ROCK_10    = 0x3A # number won't show until <= 9
    SOFT_ROCK_11    = 0x3B # number won't show until <= 9
    SOFT_ROCK_12    = 0x3C # number won't show until <= 9
    SOFT_ROCK_13    = 0x3D # number won't show until <= 9
    SOFT_ROCK_14    = 0x3E # number won't show until <= 9
    SOFT_ROCK_15    = 0x3F # number won't show until <= 9
    RED_CANDYPOP    = 0x40
    BLUE_CANDYPOP   = 0x41
    YELLOW_CANDYPOP = 0x42
    PURPLE_CANDYPOP = 0x43
    WHITE_CANDYPOP  = 0x44
    TREASURE        = 0x50
    NONE            = 0x60
# fmt: on



import logging
from pathlib import Path
from piket import NEDCENC, NEVPK
from piket.constants import VPK_SIZE, VPK
from . import to_bytes, run_tool

logger = logging.getLogger(__file__)
PARENT = NEDCENC.parent.resolve()


def decode(data: bytes | bytearray | str | Path, partial_decode: bool = False) -> bytearray:
    # handle all input types
    data = to_bytes(data)
    file = NEDCENC.parent.resolve() / "in.raw"
    logger.debug(f"Writing .raw data to '{file}'.")
    file.write_bytes(data)

    # decode .raw to .bin (includes a vpk compression block which we will extract)
    decoded_path = PARENT / "decoded.bin"
    logger.debug(f"Running nedcenc, output to '{decoded_path}'.")
    run_tool(f'"{NEDCENC}" -i "{file}" -d -o "{decoded_path}"')
    logger.debug(f"Removing '{file}'.")
    file.unlink()
    if not decoded_path.exists():
        raise Exception("nedcenc did not output a file.")

    decoded = decoded_path.read_bytes()
    logger.debug(f"Removing '{decoded_path}'.")
    decoded_path.unlink()

    if partial_decode:
        return bytearray(decoded)

    data = decompress(decoded)

    logger.info("Conversion from .raw to .bin (decompressed) complete.")
    return bytearray(data)


def decompress(decoded: bytes) -> bytes:
    # trim the decoded .bin to the vpk block (compressed level data)
    logger.debug("Trimming decoded .raw data to VPK block.")
    size = int.from_bytes(decoded[VPK_SIZE : VPK_SIZE + 2], "little")
    trimmed_path = PARENT / "trimmed.vpk"
    trimmed_path.write_bytes(decoded[VPK : VPK + size])

    # decompress the vpk block into pure level data (.bin again)
    decompressed_path = PARENT / "out.bin"
    logger.debug(f"Running nevpk, output to '{decompressed_path}'.")
    run_tool(f'"{NEVPK}" -i "{trimmed_path}" -d -o "{decompressed_path}"')
    logger.debug(f"Removing '{trimmed_path}'.")
    trimmed_path.unlink()

    if not decompressed_path.exists():
        raise Exception("nevpk did not output a file.")

    data = decompressed_path.read_bytes()
    logger.debug(f"Removing '{decompressed_path}'.")
    decompressed_path.unlink()

    return data



# python re-implementation of BlackShark's headerfix.c
# BlackShark: https://github.com/Bl4ckSh4rk
# headerfix.c: https://github.com/HunterRDev/e-ReaderCardCreator/blob/master/AC_e-Reader_Card_Creator/Decompression/External/Source%20Code/headerfix.c

HEADER_SIZE = 0x30
SIGNATURE = b"NINTENDO"


def _get_data_checksum(data: bytes) -> int:
    total = 0
    for i in range(0, len(data), 2):
        hi = data[i]
        lo = data[i + 1] if i + 1 < len(data) else 0
        total += (hi << 8) + lo
    return (~total) & 0xFFFF


def _get_header_checksum(header: bytearray) -> int:
    xor = header[0xC] ^ header[0xD] ^ header[0x10] ^ header[0x11]
    for i in range(0x26, 0x2D):
        xor ^= header[i]
    return xor & 0xFF


def _get_global_checksum(header: bytes, data: bytes) -> int:
    total = sum(header[:0x2F])
    for i in range(0, len(data), 0x30):
        xor = 0
        for b in data[i : i + 0x30]:
            xor ^= b
        total += xor
    return (~total) & 0xFF


def headerfix(input_data: bytes | bytearray) -> bytes:
    """
    Fixes the header of a decoded e-Reader card.

    Args:
        input_data (bytes or bytearray): Full contents of the e-Card.

    Returns:
        bytes: A new bytes object with updated header checksums.

    Raises:
        ValueError: If the input is invalid or corrupted.
    """
    if not isinstance(input_data, (bytes, bytearray)):
        raise TypeError("Input must be bytes or bytearray")

    if len(input_data) % HEADER_SIZE != 0 or len(input_data) <= HEADER_SIZE:
        raise ValueError("Invalid card size")

    buffer = bytearray(input_data)
    header = buffer[:HEADER_SIZE]

    if header[0x1A : 0x1A + 8] != bytearray(SIGNATURE):
        raise ValueError("Invalid e-Card signature")

    data_size = (header[0x06] << 8) + header[0x07]
    if data_size + HEADER_SIZE > len(buffer):
        raise ValueError("Data size in header exceeds buffer size")

    data = buffer[HEADER_SIZE : HEADER_SIZE + data_size]

    # checksums
    data_checksum = _get_data_checksum(data)
    header[0x13] = (data_checksum >> 8) & 0xFF
    header[0x14] = data_checksum & 0xFF

    header_checksum = _get_header_checksum(header)
    header[0x2E] = header_checksum

    global_checksum = _get_global_checksum(header, data)
    header[0x2F] = global_checksum

    return bytes(header + data)



SIZE_INFO = 0x26
VPK_SIZE = 0x51
VPK = 0x53

CARD_SET_A_PLUCKING   = "PIKMINCARD1S" # A001 -> A020
CARD_SET_B_MARCHING   = "PIKMINCARD2S" # B001 -> B020
CARD_SET_C_CONNECTING = "PIKMINCARD3S" # C001 -> C020
CARD_SETS_H_P_ALL     = "PIKMINCARDAP" # H001 -> H006 + P001 -> P003, one of each minigame (ALL)
CARD_SET_D_OLIMAR     = "PIKMINCARD1P" # D001 plucking
CARD_SET_D_PRESIDENT  = "PIKMINCARD2P" # D002 marching
CARD_SET_D_LOUIE      = "PIKMINCARD3P" # D003 connecting

PLUCKING_PIKMIN = "PIKMINPUZZLE01" .encode("ascii").ljust(16, b"\x00")
MARCHING_PIKMIN = "PIKMINPUZZLE02" .encode("ascii").ljust(16, b"\x00")
PIKMIN_OTAKARA  = "PIKMINOTAKARA"  .encode("ascii").ljust(16, b"\x00")

LEVEL_ID_LENGTH = 0x10

LEVELS_START = 0x100
LEVEL_FOOTER_LENGTH = 0x5

PLUCKING_WIDTH = 11
PLUCKING_HEIGHT = 8
PLUCKING_LAYERS = 2
PLUCKING_PIKMIN_HEADER_LENGTH = 0x5
PLUCKING_PIKMIN_LAYER_LENGTH = 0x58
PLUCKING_PIKMIN_LENGTH = PLUCKING_PIKMIN_HEADER_LENGTH + PLUCKING_PIKMIN_LAYER_LENGTH * PLUCKING_LAYERS

MARCHING_WIDTH = 32
MARCHING_HEIGHT = 11
MARCHING_PIKMIN_HEADER_LENGTH = 0x3
MARCHING_PIKMIN_LAYER_LENGTH = 0x160
MARCHING_PIKIS_LENGTH = 0x3f
MARCHING_PIKMIN_LENGTH = MARCHING_PIKMIN_HEADER_LENGTH + MARCHING_PIKMIN_LAYER_LENGTH + MARCHING_PIKIS_LENGTH

CONNECTING_WIDTH = 9
CONNECTING_HEIGHT = 6
CONNECTING_LAYERS = 3
CONNECTING_PIKMIN_HEADER_LENGTH = 0x3
CONNECTING_PIKMIN_LAYER_LENGTH = 0x36
CONNECTING_PIKMIN_LENGTH = CONNECTING_PIKMIN_HEADER_LENGTH + CONNECTING_PIKMIN_LAYER_LENGTH * CONNECTING_LAYERS

FOOTER_NORMAL = bytes([0x3E, 0x02, 0xCF, 0x00, 0xC9])
FOOTER_CONNECTING_X3 = bytes([0x3E, 0x01, 0x76, 0x18, 0xFB])



from piket import Card
from piket.treasure import Treasure
from pathlib import Path
from textwrap import dedent

_SPRITES_PATH = (Path(__file__).parent / "treasure" / "sprites").resolve()


class PrototypeDetector:
    """Helps you extract all vanilla treasure sprite data from your raw Pikmin card files."""

    def __init__(self, sprites_path: Path = _SPRITES_PATH):
        if sprites_path.is_file():
            raise Exception(f"Unexpected file at path to sprites directory: {sprites_path}")
        sprites_path.mkdir(exist_ok=True)
        self.sprites_path = sprites_path
        self._update_missing()

    def _update_missing(self):
        SPRITES = [f.name for f in self.sprites_path.glob("*.bin") if f.is_file()]
        self.missing_sprites = {
            t.name: t.value for t in list(Treasure)[:10] if f"{t.name.lower()}.bin" not in SPRITES
        }

    def get_sprites(self, path: Path | str) -> list[str]:
        """Searches all .raw card files in the path for missing vanilla treasure sprites.

        Args:
            path (Path | str): The folder to search for .raw card files.

        Returns:
            list[str]: List of names for all found treasure sprites.
        """
        self._update_missing()
        path = Path(path)
        RAWS = [f.resolve() for f in path.glob("*.raw") if f.is_file()]
        new_sprites: list[str] = []
        for raw in RAWS:
            try:
                card = Card(raw)
            except:
                print(f"Failed to open {raw.name}, continuing...")
                continue

            print(f"Opening {raw.name}...")
            if not card.treasure:
                continue

            name = card.treasure.get_name()
            key = next(
                (k for k, v in self.missing_sprites.items() if v == name and k not in new_sprites),
                None,
            )
            if not key:
                continue

            print(f"Found missing treasure sprite data for '{key}'!")
            (self.sprites_path / f"{key.lower()}.bin").write_bytes(card.treasure.encode())
            new_sprites.append(key)

        self._update_missing()
        return new_sprites


def main():
    pd = PrototypeDetector()
    if len(pd.missing_sprites) > 0:
        print(
            dedent(
                f"""
            You are currently missing the following vanilla sprites:
            {", ".join(pd.missing_sprites.keys())}

            Don't worry, though! This tool will help you get them all.

            You will need to have all the official Pikmin e-Reader cards ready in a folder.
            These look like "Pikmin Card-e+ - 12-A001 (Japan).raw".

            In total there are 72 of these cards, but if you don't have them all this script will
            still try to extract any treasure sprites it can find.

            Enter the path to a folder with all your .raw card files here:"""
            )
        )

        while True:
            input_path = input()

            path = Path(input_path).resolve()
            if not path.is_dir():
                print("\nThat path was not recognised as a folder, please try again:")

            else:
                found = pd.get_sprites(path)
                if len(found) == 0:
                    print(
                        "I couldn't find any new treasure sprites in the provided directory, "
                        "please try again:"
                    )

                else:
                    print(
                        "Good news! I found the following treasure sprites and stored them:\n"
                        f"{", ".join(found)}\n"
                    )
                    if len(pd.missing_sprites) == 0:
                        print("You should now be able to use all vanilla treasure sprites, enjoy!")
                        exit()

                    print(
                        f"The following treasure sprites are still missing:"
                        f"{", ".join(pd.missing_sprites.keys())}"
                    )
                    print("If you'd like to try a new folder, enter it here:")

    else:
        print("All treasure sprites were found, good to go! :)")


if __name__ == "__main__":
    main()



import sys
import platform
import importlib.resources as resources
from pathlib import Path

_TOOLS = {
    "windows": {
        "libnedclib": "libnedclib.dll",
        "nedcenc": "nedcenc.exe",
        "nevpk": "nevpk.exe",
    },
    "macos": {
        "libnedclib": "libnedclib.dylib",
        "nedcenc": "nedcenc",
        "nevpk": "nevpk",
    },
    "linux": {
        "libnedclib": "libnedclib.so",
        "nedcenc": "nedcenc",
        "nevpk": "nevpk",
    },
}


def get_machine():
    os_name = sys.platform
    machine = platform.machine().lower()

    # Normalize OS name
    if os_name.startswith("linux"):
        os_name = "linux"
    elif os_name == "darwin":
        os_name = "macos"
    elif os_name in ("win32", "cygwin", "msys"):
        os_name = "windows"

    # Normalize architecture
    if machine in ("amd64", "x86_64"):
        arch = "amd64"
    elif machine in ("aarch64", "arm64"):
        arch = "arm64"
    else:
        arch = machine

    return f"{os_name}-{arch}"


# validate platform os support
machine = get_machine()
plat = machine[: machine.index("-")]
platform_tools = _TOOLS.get(plat)
if not platform_tools:
    raise OSError(f"Piket currently does not support: {machine}")

# resolve tool paths and expose them
_TOOL_PATHS: dict[str, Path] = {}
for tool, filename in platform_tools.items():
    try:
        with resources.path(f"piket.bin.{machine}", filename) as p:
            if not Path(p).exists():
                raise FileNotFoundError(f"Missing required tool: {tool}")
            _TOOL_PATHS[tool] = p
    except Exception as e:
        raise ImportError(f"Error loading binary '{filename}': {e}")

NEDCENC = _TOOL_PATHS["nedcenc"]
NEVPK = _TOOL_PATHS["nevpk"]

# expose methods and classes
from .util import decode, encode, get_id
from .card import Card
from .treasure import Treasure
from . import connecting_pikmin as ConnectingPikmin
from . import plucking_pikmin as PluckingPikmin
from . import marching_pikmin as MarchingPikmin

__all__ = [
    # tools
    "NEDCENC",
    "NEVPK",
    # direct methods
    "decode",
    "encode",
    "get_id",
    # primary classes
    "Card",
    "Treasure",
    # level classes
    "ConnectingPikmin",
    "PluckingPikmin",
    "MarchingPikmin",
]



import logging
from pathlib import Path
from .decode import decode

logger = logging.getLogger(__file__)


def get_id(data: bytes | bytearray | str | Path) -> bytes:
    decoded = decode(data, partial_decode=True)
    id = decoded[-16:]
    return bytes(id)



from enum import IntEnum


# fmt: off
class Piki(IntEnum):
    NONE            = 0
    RED             = 1
    BLUE            = 2
    YELLOW          = 3
    RED_PLUCKED     = 5
    BLUE_PLUCKED    = 6
    YELLOW_PLUCKED  = 7
# fmt: on



import logging
import subprocess
from pathlib import Path

logger = logging.getLogger(__file__)


def to_bytes(data: bytes | bytearray | str | Path) -> bytes:
    if isinstance(data, (bytes, bytearray)):
        return bytes(bytearray(data))
    elif isinstance(data, (str, Path)):
        with open(data, "rb") as f:
            return f.read()
    else:
        raise TypeError("Expected bytes, bytearray, or file path (str/Path).")


def run_tool(command: str):
    result = subprocess.run(
        command,
        capture_output=True,
        text=True,
        shell=True,  # kept getting FileNotFound for long filenames w/ shell=False + commmand list[]
    )

    if result.stdout:
        for line in result.stdout.splitlines():
            logger.debug(line)

    if result.stderr:
        for line in result.stderr.splitlines():
            logger.error(line)

    if result.returncode != 0:
        raise subprocess.CalledProcessError(result.returncode, command)


# expose methods directly
from .decode import decode, decompress
from .encode import encode
from .get_id import get_id

__all__ = ["decode", "decompress", "encode", "get_id"]



from pathlib import Path
from .util import to_bytes, decode, decompress, encode
from .constants import *
from pathlib import Path
from . import (
    plucking_pikmin as P,
    connecting_pikmin as C,
    marching_pikmin as M,
)
from .base.level_base import LevelBase
from .treasure import Treasure, TreasureSprite
import logging

logger = logging.getLogger(__file__)


class Card:

    levels: list[LevelBase]
    treasure: TreasureSprite | None
    raw: bytes
    decoded: bytes
    leveldata: bytes
    footer: bytes
    id: str

    def __init__(self, card: bytes | bytearray | str | Path | None):
        """Creates a new Card object from .raw bytes."""
        self.levels: list[LevelBase] = []
        self.treasure: TreasureSprite | None = None
        if card is None:
            self.footer = FOOTER_NORMAL

        else:
            in_bytes = to_bytes(card)
            in_footer = in_bytes[-5:]
            # determine type of input bytes (card.raw, card_decoded.bin, card_decompressed.bin)
            if in_footer == FOOTER_NORMAL or in_footer == FOOTER_CONNECTING_X3:
                # in bytes are decompressed level data
                # unfortunately we will have to infer a level ID based on contents
                self.leveldata = in_bytes
                raise NotImplementedError(
                    "Automatic inference of ID based on level data contents alone "
                    "has not been implemented yet."
                )

            elif in_bytes[0x1A : 0x1A + 8] == b"NINTENDO":
                # in bytes are decoded card data which needs to be decompressed
                self.decoded = in_bytes
                self.leveldata = decompress(in_bytes)

            else:
                # raw card file
                self.raw = in_bytes
                self.decoded = decode(in_bytes, partial_decode=True)
                self.leveldata = decompress(self.decoded)

            if self.decoded:
                self.id = self.decoded[-16:].decode("ascii").replace("\x00", "")

            self.footer = self.leveldata[-5:]

            # check footer offset
            # only applies to any card that isn't 3x Connecting Pikmin, since
            # they skip the initial 0x100 padding
            if self.leveldata[0] == 0xC3:
                # only nintendo and God knows why the offset is +0x100
                footer_offset = int.from_bytes(self.leveldata[1:3], "little") - 0x100
                if len(self.leveldata) != footer_offset + 5:
                    raise ValueError(
                        f"Decoded+decompressed card contained a footer offset that did not match. "
                        f"Expected length {hex(footer_offset+5)}, got {hex(len(self.leveldata))}."
                    )

            if self.id == CARD_SET_A_PLUCKING or self.id == CARD_SET_D_OLIMAR:
                """
                Set-A cards and also Set-D001/Olimar. Contains:
                - 3 x Plucking
                """
                start = LEVELS_START + LEVEL_ID_LENGTH
                full_level = LEVEL_ID_LENGTH + PLUCKING_PIKMIN_LENGTH
                for i in range(3):
                    end = min(start + full_level, len(self.leveldata) - LEVEL_FOOTER_LENGTH)
                    level_data = self.leveldata[start:end]
                    level = P.Level.from_bytes(level_data)
                    self.levels.append(level)
                    start += full_level

            elif self.id == CARD_SET_B_MARCHING or self.id == CARD_SET_D_PRESIDENT:
                """
                Set-B cards and also Set-D002/President. Contains:
                - 3 x Marching
                """
                start = LEVELS_START * 2 + LEVEL_ID_LENGTH
                full_level = LEVEL_ID_LENGTH + MARCHING_PIKMIN_LENGTH
                for i in range(3):
                    end = start + full_level
                    level_data = self.leveldata[start:end]
                    level = M.Level.from_bytes(level_data)
                    self.levels.append(level)
                    start += full_level

            elif self.id == CARD_SET_C_CONNECTING or self.id == CARD_SET_D_LOUIE:
                """
                Set-C cards and also Set-D003/Louie. Contains:
                - 3 x Connecting
                """
                for i in range(3):
                    start = i * 0x100
                    end = min((i + 1) * 0x100, len(self.leveldata) - LEVEL_FOOTER_LENGTH)
                    level_data = self.leveldata[start:end]
                    level = C.Level.from_bytes(level_data)
                    self.levels.append(level)

            elif self.id == CARD_SETS_H_P_ALL:
                """
                Promotional cards [H001 -> H006] + [P001 -> P003]. Contains:
                - 1 x Plucking
                - 1 x Marching
                - 1 x Connecting
                """
                start = LEVELS_START + LEVEL_ID_LENGTH
                end = 0x200
                level_data = self.leveldata[start:end]
                level = P.Level.from_bytes(level_data)
                self.levels.append(level)
                # Marching Pikmin
                start = 0x200 + LEVEL_ID_LENGTH
                end = 0x400
                level_data = self.leveldata[start:end]
                level = M.Level.from_bytes(level_data)
                self.levels.append(level)
                # Connecting Pikmin
                start = 0x400
                end = len(self.leveldata) - LEVEL_FOOTER_LENGTH
                level_data = self.leveldata[start:end]
                level = C.Level.from_bytes(level_data)
                self.levels.append(level)

            else:
                raise ValueError(f"Card data contains an unrecognised ID: '{self.id}'.")

            # check for any treasure sprite data
            self._decode_treasures()

    def _decode_treasures(self) -> bool:
        """Decodes this card's PIKMINOTAKARA block if it contains one.

        Returns:
            bool: True if succeeds, False if there is no data to decode.
        """
        if not self.leveldata:
            raise Exception(
                "Unable to decode treasures without decoded and " "decompressed level data."
            )

        # make sure the block at least big enough to validate header
        if len(self.leveldata) > 0x790:
            data = self.leveldata[0x780:]
            # make sure the block starts with PIKMINOTAKARA header
            if self.leveldata[0x780:0x790] == PIKMIN_OTAKARA:
                self.treasure = TreasureSprite.from_bytes(data)
                return True

            # there shouldn't be any other blocks at 0x780, so raise exception for unknown block
            else:
                otakara_decoded = PIKMIN_OTAKARA.decode("ascii").replace("\x00", ".")

                # try to decode the block header if possible
                found_decoded = "Failed to decode!"
                try:
                    found_decoded = data[:16].decode("ascii")
                except:
                    pass

                raise ValueError(
                    f"Unrecognised header at Treasure Sprite Data offset:\n"
                    f"Expected 0x{PIKMIN_OTAKARA.hex()} ({otakara_decoded})\n"
                    f"Got 0x{data[:16].hex()} ({found_decoded})"
                )

        return False

    def set_treasure(self, treasure: Treasure):
        """Sets the treasure binary data automatically.
        IMPORTANT: This only works if you have already extracted your treasure sprites
        using the Prototype Detector. Run `python -m piket.pd` to confirm.
        """
        self.treasure = TreasureSprite.from_bytes(treasure.data)

    def encode(self, partial_encode: bool = False, raw_level: bool = False) -> bytes:
        new_leveldata = bytearray()
        for i, level in enumerate(self.levels):
            if i > 2:
                raise ValueError(
                    f"Cards are only designed to contain 3 levels, got {len(self.levels)}."
                )
            # first level always starts different (Connecting Pikmin goes straight to level data)
            if i == 0:
                if isinstance(level, P.Level):
                    # always 0x100 padding before first Plucking Pikmin level
                    new_leveldata.extend(self.leveldata[:0x100])

                elif isinstance(level, M.Level):
                    # always 0x200 padding before first Marching Pikmin level
                    new_leveldata.extend(self.leveldata[:0x200])

            assert isinstance(level, (P.Level, M.Level, C.Level))
            level_bytes = level.to_bytes()

            # size and padding is different for each level
            expected_size = 0
            padding = 0
            if isinstance(level, P.Level):
                expected_size = 0xC5
                # pad to 0x100 if its the first level in a non-Set-A and non-Set-D-Olimar card
                if i == 0 and self.id not in [CARD_SET_A_PLUCKING, CARD_SET_D_OLIMAR]:
                    padding = 0x100

            elif isinstance(level, M.Level):
                expected_size = 0x1B2
                # pad to 0x200 if its the second level in a non-Set-B and non-Set-D-President card
                if i == 1 and self.id not in [CARD_SET_B_MARCHING, CARD_SET_D_PRESIDENT]:
                    padding = 0x200

            else:
                expected_size = 0xA5
                # non-last Connecting Pikmin levels are always padded to 0x100
                if i < 2:
                    padding = 0x100
                # if its the last level, add no padding

            # validate level sizes
            if len(level_bytes) < expected_size:
                logger.warning(
                    f"Level '{level.__class__.__module__}' size less than expected "
                    f"{hex(expected_size)}, got {hex(len(level_bytes))}.\n"
                    f"The level will be padded to {hex(expected_size)}."
                )
                level_bytes = level_bytes.ljust(expected_size, b"\x00")

            elif len(level_bytes) > expected_size:
                raise ValueError(
                    f"Level '{level.__class__.__module__}' size larger than expected "
                    f"{hex(expected_size)}, got {hex(len(level_bytes))}"
                )

            # pad further if required
            if padding > 0:
                level_bytes = level_bytes.ljust(padding, b"\x00")

            # add level bytes
            new_leveldata.extend(level_bytes)

        # add PIKMINOTAKARA treasure data if necessary
        if self.treasure:
            # always pad to 0x780 for treasure data
            new_leveldata = new_leveldata.ljust(0x780, b"\x00")
            new_leveldata.extend(self.treasure.encode())

        # add footer
        new_leveldata.extend(self.footer)

        # return early if user wants raw level data bytes
        if raw_level:
            return bytes(new_leveldata)

        # rebuild raw from existing decoded data
        elif self.decoded:
            out = encode(new_leveldata, self.decoded, partial_encode)

        else:
            # reconstruct raw data from inferred data
            raise NotImplementedError(
                "Automatic decoded data construction from just level data is incomplete."
            )

        # only update self.raw if it was a full encode
        if not partial_encode:
            self.raw = out

        # return encoded (full or partial) bytes
        return bytes(out)



from enum import IntEnum


# fmt: off
class Tile(IntEnum):
    NONE                = 0x00
    GRASS               = 0x01
    ROCK                = 0x02
    FIRE                = 0x03
    WATER               = 0x04
    ELECTRICITY_NODE    = 0x05
    ELECTRICTY          = 0x06
    SLIME_GRASS         = 0x9B # likely a sprite glitch
# fmt: on



from enum import IntEnum


# fmt: off
class Tile(IntEnum):
    ALL = 0
    HORIZONTAL      = 1
    VERTICAL        = 2
    TOP_LEFT        = 3
    TOP_RIGHT       = 4
    BOTTOM_LEFT     = 5
    BOTTOM_RIGHT    = 6
    NONE            = 7
# fmt: on



from enum import Enum
from pathlib import Path


class Treasure(Enum):
    # fmt: off
    TEAR_STONE            = "ティア・ストーン"
    PINK_MENACE           = "オーバーキル・レッド"
    SCIENCE_PROJECT       = "ラブリー押し葉"
    AQUATIC_MINE          = "アクアティックマイン"
    CHOCOLATE_CUSHION     = "チョコベッド"
    SUNSEED_BERRY         = "バクレツイチゴ"
    COLOSSAL_FOSSIL       = "古代生物の化石"
    HEAVY_DUTY_MAGNETIZER = "超磁性体マグナイザー"
    DANGER_CHIME          = "ミスティックドーム"
    THE_KEY               = "あのカギ"

    # custom treasures: feel free to make a PR with your own creations
    FUEL_RESERVOIR        = "Fuel Reservoir"
    # fmt: on

    @property
    def data(self) -> bytes:
        path = Path(__file__).parent / "sprites" / (self.name.lower() + ".bin")
        if not path.is_file():
            raise FileNotFoundError(
                f"Treasure sprite data for '{self.name}' was not found.\n\n"
                "If this is a vanilla treasure, then that likely means\nyou neeed to run: "
                "`python -m piket.pd` and follow the prompts to extract the sprites."
            )

        return path.read_bytes()



from PIL import Image
import struct
from pathlib import Path
from io import BytesIO
from typing import Self, cast
from piket.constants import PIKMIN_OTAKARA
import logging

logger = logging.getLogger(__file__)


class TreasureSprite:
    """Treasure Sprite data class for decoding and encoding GBA sprite data

    Follow this guide for creating custom treasures:
    1. Use `treasure.export_palette("out.pal")` to get the (*.pal) file
    2. Open YY-CHR and go to "Palette" > "Open RGB Palette (*.pal)..." and select the exported palette
    3. Set "Format" to "4BPP GBA"
    6. Set "Pattern" to "32x32(B)" for the large sprite, or "FC/NES x8" for the small sprite
    7. Create your sprite and modify the palette to your liking (shared by both small and large sprites)
    8. Export palette via "Palette" > "Save RGB Palette (*.pal)..."
    9. Export sprites via "File" > "Save Bitmap..."
    10. Use `treasure.import_palette("new_palette.pal")`
    11. Use `treasure.import_small("new_16x16.bmp")`
    12. Use `treasure.import_large("new_32x32.bmp")`

    NOTE: The palette is shared between the small and large sprites.
    """

    def __init__(
        self,
        header: bytes = bytes(2),
        palette: bytes = bytes(0x20),
        sprites: bytes = bytes(0x280),
        name: bytes = bytes(0x32),
    ):
        def check_size(b: bytes, expected: int, name: str):
            if len(b) != expected:
                raise ValueError(
                    f"Invalid {name} size for PIKMINOTAKARA block: "
                    f"expected {hex(expected)}, got: {hex(len(b))}"
                )

        check_size(header, 2, "header")
        check_size(palette, 0x20, "palette")
        check_size(sprites, 0x280, "sprites")
        check_size(name, 0x32, "name")

        self.header = header
        self.palette = bytearray(palette)
        self.sprites = bytearray(sprites)
        self.name = name

    @classmethod
    def from_bytes(cls, data: bytes | bytearray) -> Self:
        """Converts the block data into a TreasureSprite class with useful methods."""
        MAGIC_SIZE = 16
        UNK_HEADER_SIZE = 2
        PALETTE = MAGIC_SIZE + UNK_HEADER_SIZE
        PALETTE_SIZE = 0x20
        SPRITE_SML = PALETTE + PALETTE_SIZE
        SPRITE_SML_SIZE = 0x80
        SPRITE_LRG = SPRITE_SML + SPRITE_SML_SIZE
        SPRITE_LRG_SIZE = 0x200
        NAME = SPRITE_LRG + SPRITE_LRG_SIZE
        NAME_SIZE = 0x32
        END = NAME + NAME_SIZE
        if len(data) < END:
            raise ValueError(
                f"Treasure sprite data block smaller than expected: "
                f"got {hex(len(data))} but expected {hex(END)}"
            )
        header = data[MAGIC_SIZE:PALETTE]
        palette = data[PALETTE:SPRITE_SML]
        sprites = data[SPRITE_SML:NAME]
        name = data[NAME:END]
        return cls(header, palette, sprites, name)

    def encode(self) -> bytes:
        """Returns the full bytes for this PIKMINOTAKRA block. Used by the Card encoder."""
        out = bytearray()
        out.extend(PIKMIN_OTAKARA)
        out.extend(self.header)
        out.extend(self.palette)
        out.extend(self.sprites)
        out.extend(self.name)

        # validate size
        if len(out) < 0x2E4:
            logger.warning(
                f"PIKMINOTAKARA size less than expected 0x2e4, got {hex(len(out))}.\n"
                f"The block will be padded to 0x2e4."
            )
            out = out.ljust(0x2E4, b"\x00")

        elif len(out) > 0x2E4:
            raise ValueError(
                f"Card treasure sprite data is too large. "
                f"Expected <= 0x2e4, got {hex(len(out))}"
            )

        return bytes(out)

    # --- palette conversion ---
    @staticmethod
    def gba_to_rgb(col16: int) -> tuple[int, int, int]:
        r = (col16 & 0x1F) << 3
        g = ((col16 >> 5) & 0x1F) << 3
        b = ((col16 >> 10) & 0x1F) << 3
        return (r, g, b)

    @staticmethod
    def rgb_to_gba(rgb: tuple[int, int, int]):
        r, g, b = rgb
        return (r >> 3) | ((g >> 3) << 5) | ((b >> 3) << 10)

    def get_palette_as_rgb(self) -> list[tuple[int, int, int]]:
        """Gets this treasure's palette as a list of 16 RGB values like (R, G, B)."""
        return [
            self.gba_to_rgb(struct.unpack("<H", self.palette[i : i + 2])[0])
            for i in range(0, 0x20, 2)
        ]

    def set_palette_from_rgb(self, colours: list[tuple[int, int, int]]):
        """Sets this treasure's palette from a list of 16 RGB values like (R, G, B)."""
        out = bytearray()
        for rgb in colours[:16]:
            # extends with little endian unsigned short
            out += struct.pack("<H", self.rgb_to_gba(rgb))
        self.palette = bytes(out)

    # --- tile conversion ---
    @staticmethod
    def decode_4bpp_tile(tile_bytes: bytes) -> list[list[int]]:
        """Decodes a 32 byte 8x8 4BPP tile to 8x8 list."""
        tile = [[0] * 8 for _ in range(8)]

        for y in range(8):
            for x in range(4):
                byte = tile_bytes[y * 4 + x]
                # split byte into 2 pixels
                tile[y][x * 2] = byte & 0xF
                tile[y][x * 2 + 1] = byte >> 4

        return tile

    @staticmethod
    def encode_4bpp_tile(tile: list[list[int]]) -> bytes:
        """Encodes an 8x8 list into a 32 byte 4BPP tile."""
        out = bytearray()
        for y in range(8):
            for x in range(0, 8, 2):
                # pixel A in low 4 bits
                # pixel B in high 4 bits
                out.append(tile[y][x] | (tile[y][x + 1] << 4))

        return bytes(out)

    # --- sprite encoder/decode ---
    def sprite_to_image(
        self, sprite_bytes: bytes, layout: list[list[int]], tile_count: int, width: int, height: int
    ) -> Image.Image:
        """Converts the sprite binary data to an Image, using the palette data."""

        # get tiles from sprite data
        tiles = [
            self.decode_4bpp_tile(sprite_bytes[i * 32 : (i + 1) * 32]) for i in range(tile_count)
        ]

        # new image in palette-based mode
        img = Image.new("P", (width, height))

        # get palette data
        palette = self.get_palette_as_rgb()
        flattened = sum(palette, ())
        img.putpalette(flattened)

        # individually place pixels
        for ty, row in enumerate(layout):
            for tx, tile_idx in enumerate(row):
                tile = tiles[tile_idx]
                for y in range(8):
                    for x in range(8):
                        img.putpixel((tx * 8 + x, ty * 8 + y), tile[y][x])

        return img

    def image_to_sprite(self, img: Image.Image, layout: list[list[int]], tile_count: int):
        """Converts an Image into sprite binary data."""
        img = img.convert("P")
        sprite_bytes = bytearray()

        # some gba formats have different tile layouts
        # in our case, the 16x16 and 32x32 are slightly different
        for ty, row in enumerate(layout):
            for tx, _ in enumerate(row):
                # extract 8x8 tile starting at (tx * 8, ty * 8)
                tile = [
                    [cast(int, img.getpixel((tx * 8 + x, ty * 8 + y))) for x in range(8)]
                    for y in range(8)
                ]
                sprite_bytes.extend(self.encode_4bpp_tile(tile))

        # limit total sprite bytes to tile count
        return bytes(sprite_bytes[: tile_count * 32])

    # --- public methods ---
    def export_small(self, path: str | Path | None = None) -> Image.Image:
        layout = [[0, 1], [2, 3]]
        small_sprite = self.sprites[:0x80]
        img = self.sprite_to_image(small_sprite, layout, 4, 16, 16)
        if isinstance(path, str | Path):
            img.save(path)

        return img

    def export_large(self, path: str | Path | None = None) -> Image.Image:
        layout = [
            [0, 1, 2, 3],
            [4, 5, 6, 7],
            [8, 9, 10, 11],
            [12, 13, 14, 15],
        ]
        large_sprite = self.sprites[0x80:0x280]
        img = self.sprite_to_image(large_sprite, layout, 16, 32, 32)
        if isinstance(path, str | Path):
            img.save(path)

        return img

    def import_small(self, sprite: str | Path | BytesIO | Image.Image):
        """Encodes a 16x16 8bpp indexed Bitmap/PNG-8 file as sprite data and adds to the card.\n
        Highly recommended guide:
        1. Use `treasure.export_palette("out.pal")` to get the (*.pal) file
        2. Open YY-CHR and go to "Palette" > "Open RGB Palette (*.pal)..." and select the exported palette
        3. Set "Format" to "4BPP GBA"
        4. Set "Pattern" to "FC/NES x8" or "32x32(A)"
        5. Create your 16x16 sprite, palette can be edited but will be shared with 32x32 sprite
        8. Export palette via "Palette" > "Save RGB Palette (*.pal)..."
        9. Export sprite via "File" > "Save Bitmap..."
        10. Use `treasure.import_palette("new_pal.pal")`
        11. Use `treasure.import_small("new_16x16.bmp")`

        NOTE: The palette is shared between the small and large sprites.
        """
        layout = [[0, 1], [2, 3]]
        if isinstance(sprite, Image.Image):
            img = sprite
        else:
            img = Image.open(sprite)

        self.sprites[:0x80] = self.image_to_sprite(img, layout, 4)

    def import_large(self, sprite: str | Path | BytesIO | Image.Image):
        layout = [
            [0, 1, 2, 3],
            [4, 5, 6, 7],
            [8, 9, 10, 11],
            [12, 13, 14, 15],
        ]
        if isinstance(sprite, Image.Image):
            img = sprite
        else:
            img = Image.open(sprite)

        self.sprites[0x80:0x280] = self.image_to_sprite(img, layout, 16)

    def import_palette(self, palette: str | Path | bytes):
        """Imports 16-bit palette data from (*.pal) file."""
        if isinstance(palette, str | Path):
            p = Path(palette)
            if not p.is_file():
                raise FileNotFoundError(
                    f"Could not import palette from file that does not exist: '{palette}'."
                )
            data = p.read_bytes()

        else:
            data = palette

        if len(data) != 0x200:
            raise ValueError(
                f"Import palette requires a 16-bit colour palette (0x200 size), "
                f"got size {hex(len(data))}."
            )

        self.palette = data[:0x20]

    def export_palette(self, path: str | Path | None = None) -> bytes:
        """Exports 16-bit palette data to (*.pal) file."""
        if isinstance(path, str | Path):
            # pad to 0x200
            palette = bytearray(self.palette).ljust(0x200, b"\x00")
            Path(path).write_bytes(palette)

        return self.palette

    def get_name(self) -> str:
        """Gets the SHIFT-JIS encoded name of this treasure, stripped of null bytes."""
        return self.name.decode("shift-jis").strip("\x00")

    def set_name(self, name: str):
        """Sets the SHIFT-JIS encoded name of this treasure. Max length 48 bytes."""
        encoded = name.encode("shift-jis")
        if len(encoded) > 48:
            raise ValueError(
                f"Treasure name must <= 48 bytes (SHIFT-JIS encoded), got {len(encoded)}"
            )

        self.name = encoded.ljust(0x32, b"\x00")



from abc import ABC, abstractmethod
from typing import Self


class LevelBase(ABC):
    """Abstract base class for all level types."""

    index: int
    width: int
    height: int
    layers: int
    tiles: list[bytearray]
    raw: bytes

    def __init__(
        self,
        index: int,
        width: int,
        height: int,
        layers: int = 2,
        tiles: list[bytearray] | None = None,
        raw: bytes = bytes(0),
    ):
        """Initialise a level with width, height, and layered tiles.

        Args:
            index (int): Index of this level in its game mode.
            width (int): Number of tiles horizontally.
            height (int): Number of tiles vertically.
            layers (int, optional): Number of tile layers. Defaults to 2.
            tiles (list[bytearray] | None, optional): Pre-filled tile data. Defaults to None.
            raw (bytes | None, optional): Original decoded level data. Defaults to 0.
        """
        self.index = index
        self.width = width
        self.height = height
        self.layers = layers
        self.tiles = tiles or [bytearray(width * height) * layers]
        self.raw = raw

    @classmethod
    @abstractmethod
    def from_bytes(cls, level: bytes | bytearray) -> Self:
        """Create a level from decoded level data bytes.

        Args:
            level (bytes | bytearray): Decoded level data bytes

        Returns:
            LevelBase: A new level instance.
        """
        pass

    def _to_bytes(self, raw: bytearray = bytearray()) -> bytes:
        """Extends a bytes object with this level's tile layers.

        Returns:
            bytes: Level data as bytes.
        """
        for i in range(self.layers):
            raw.extend(self.tiles[i])
        return raw

    def _get_tile(self, x: int, y: int, layer: int) -> int:
        """Gets the tile at (x, y, layer).

        Returns:
            int: The exact tile value from bytes.
        """
        if not (0 <= layer < len(self.tiles)):
            raise IndexError(f"Invalid layer {layer}")
        if not (0 <= x < self.width and 0 <= y < self.height):
            raise IndexError(f"Invalid position ({x}, {y})")
        return self.tiles[layer][y * self.width + x]

    def _set_tile(self, x: int, y: int, tile: int, layer: int):
        """Sets the tile at (x, y, layer) to an int value."""
        self.tiles[layer][y * self.width + x] = tile

    def _set_tiles(self, x: int, y: int, w: int, h: int, tile: int, layer: int):
        """Sets the tiles at (x, y, w, h, layer) to an int value."""
        for ix in range(x, x + w):
            for iy in range(y, y + h):
                # * must use LevelBase._set_tile() because self may not be a LevelBase
                LevelBase._set_tile(self, ix, iy, tile, layer)



from .tile import Tile
from .piki import Piki
from piket.constants import (
    MARCHING_PIKMIN,
    MARCHING_PIKMIN_HEADER_LENGTH as HEADER_LEN,
    MARCHING_PIKMIN_LAYER_LENGTH as LAYER_LEN,
    MARCHING_WIDTH as WIDTH,
    MARCHING_HEIGHT as HEIGHT,
    MARCHING_PIKIS_LENGTH as PIKIS_LEN,
)
from piket.base.level_base import LevelBase
from typing import Self
import logging

logger = logging.getLogger(__file__)


class Level(LevelBase):
    def __init__(
        self,
        index: int,
        tiles: bytearray = bytearray(WIDTH * HEIGHT),
        pikis: list[tuple[int, int, int]] = [],
        raw: bytes | bytearray = bytearray(0),
        unk01: int = 0,
        use_custom_treasure: bool = False,
    ):
        super().__init__(index, WIDTH, HEIGHT, 1, [tiles], raw)
        self.pikis = pikis
        self.unk01 = unk01
        self.use_custom_treasure = use_custom_treasure

    @classmethod
    def from_bytes(cls, level: bytes | bytearray) -> Self:
        level = bytearray(level)
        index = level[0]
        unk01 = level[1]
        use_custom_treasure = level[2] > 0
        layers = level[HEADER_LEN:]
        tiles = layers[:LAYER_LEN]
        pikis_bytes = layers[LAYER_LEN : LAYER_LEN + PIKIS_LEN]
        pikis: list[tuple[int, int, int]] = []
        for i in range(0, 33, 3):  # 11 pikis total
            x = pikis_bytes[i]
            y = pikis_bytes[i + 1]
            # break if x,y = 0,0 (same as game)
            if x == 0 and y == 0:
                break

            piki_type = pikis_bytes[i + 2]
            # skip if outside range (causes game crash)
            if piki_type > 4:
                logger.warning(
                    f"Marching Pikmin level data contains unrecognised Piki "
                    "type '{piki_type}'. This will be skipped to avoid game crashes."
                )
                continue

            pikis.append((x, y, piki_type))

        return cls(index, tiles, pikis, level, unk01, use_custom_treasure)

    def to_bytes(self) -> bytes:
        raw = bytearray()
        raw.extend(MARCHING_PIKMIN)
        raw.append(self.index)
        raw.append(self.unk01)
        raw.append(0xFF if self.use_custom_treasure else 0)
        super()._to_bytes(raw)
        for i in range(0, 0x3F, 3):
            if len(self.pikis) > i // 3:
                piki = self.pikis[i // 3]
                raw.extend(piki)
            else:
                # note that the vanilla levels seem to have a random number of 0x00 0x00 0x0F
                # the game stops reading pikis as soon as it sees x=0 && y=0
                # so it doesn't matter, just extend with zeroes
                # this does unfortunately mean we will not get 1:1 input+output from vanilla levels
                raw.extend([0, 0, 0])
        return bytes(raw)

    def get_tile(self, x: int, y: int) -> Tile:
        """Gets the Tile at (x, y)."""
        value = super()._get_tile(x, y, 0)
        if value not in Tile._value2member_map_:
            raise ValueError(f"Unknown Tile with value {value}")
        return Tile(value)

    def set_tile(self, x: int, y: int, tile: Tile):
        """Sets the Tile at (x, y)."""
        super()._set_tile(x, y, tile.value, 0)

    def set_tiles(self, x: int, y: int, w: int, h: int, tile: Tile):
        """Sets the Tiles from (x, y) to (w, h)."""
        super()._set_tiles(x, y, w, h, tile.value, 0)

    def add_piki(self, x: int, y: int, piki: Piki):
        """Places a Piki at (x, y). Cannot be (0, 0)."""
        if x == 0 and y == 0:
            raise ValueError(f"Marching Pikmin does not support Pikmin at (0, 0).")
        if len(self.pikis) >= 20:
            raise ValueError("Marching Pikmin is limited to 20 Pikmin.")
        self.pikis.append((x, y, piki.value))

    def clear_all(self):
        """Sets all Tiles to value 0x60 and removes all Pikmin."""
        for i in range(self.layers):
            super()._set_tiles(0, 0, self.width, self.height, 0x60, i)
        self.clear_pikis()

    def clear_pikis(self):
        """Removes all Pikis currently in the level."""
        self.pikis.clear()
